@@ -3,7 +3,7 @@ package DBIx::BatchChunker;
 our $AUTHORITY = 'cpan:GSG';
 # ABSTRACT: Run large database changes safely
 use version;
-our $VERSION = 'v0.940.1'; # VERSION
+our $VERSION = 'v0.940.2'; # VERSION
 
 use Moo;
 use MooX::StrictConstructor;
@@ -21,6 +21,8 @@ use Scalar::Util            qw( blessed weaken );
 use Term::ProgressBar 2.14;                          # with silent option
 use Time::HiRes             qw( time sleep );
 
+use DBIx::BatchChunker::LoopState;
+
 # Don't export the above, but don't conflict with StrictConstructor, either
 use namespace::clean -except => [qw< new meta >];
 
@@ -160,6 +162,11 @@ our $DB_MAX_ID = Data::Float::max_integer;  # used for progress_past_max
 #pod constructor to enable features like L<max ID recalculation|/process_past_max> or
 #pod L<chunk resizing|/min_chunk_percent>.
 #pod
+#pod If you're not going to include any min/max statements for L</calculate_ranges>, you will
+#pod need to set L</min_id> and L</max_id> yourself, either in the constructor or before the
+#pod L</execute> call.  Using L</construct_and_execute> is also not an option in this case, as
+#pod this tries to call L</calculate_ranges> without a way to do so.
+#pod
 #pod =head3 TL;DR Version
 #pod
 #pod     $stmt                             = Active DBI Processing
@@ -211,11 +218,10 @@ has rsc => (
 #pod     retry_handler = Coderef that returns true to continue to retry or false to re-throw
 #pod                     the last exception
 #pod
-#pod The default is to use DBIC's built-in retry options, the same way L<DBIx::Class::Storage::DBI/dbh_do>
-#pod does it, which will retry once if the DB connection was disconnected.  If you specify any
-#pod options, even a blank hashref, BatchChunker will fill in a default C<max_attempts> of 10,
-#pod and an always-true C<retry_handler>.  This is similar to L<DBIx::Connector::Retry>'s
-#pod defaults.
+#pod The default is to let the DBIC storage engine handle its own protection, which will retry
+#pod once if the DB connection was disconnected.  If you specify any options, even a blank
+#pod hashref, BatchChunker will fill in a default C<max_attempts> of 10, and an always-true
+#pod C<retry_handler>.  This is similar to L<DBIx::Connector::Retry>'s defaults.
 #pod
 #pod Under the hood, these are options that are passed to the as-yet-undocumented
 #pod L<DBIx::Class::Storage::BlockRunner>.  The C<retry_handler> has access to the same
@@ -225,27 +231,23 @@ has rsc => (
 #pod =cut
 
 has dbic_retry_opts => (
-    is       => 'ro',
-    isa      => HashRef,
-    required => 0,
-    lazy     => 1,
-    builder  => 1,
+    is        => 'ro',
+    isa       => HashRef,
+    required  => 0,
+    predicate => '_has_dbic_retry_opts',
 );
 
-sub _build_dbic_retry_opts {
-    my $self = shift;
-
-    return {
-        # the default from DBIC
-        retry_handler => sub {
-            $_[0]->failed_attempt_count == 1 && !$_[0]->storage->connected
-        },
-    };
-}
-
 sub _dbic_block_runner {
     my ($self, $method, $coderef) = @_;
 
+    my $storage = $self->dbic_storage;
+
+    # Block running disabled
+    unless ($self->_has_dbic_retry_opts) {
+        return $storage->txn_do($coderef) if $method eq 'txn';
+        return $storage->dbh_do($coderef);
+    }
+
     # A very light wrapper around BlockRunner.  No need to load BlockRunner, since DBIC
     # loads it in before us if we're using this method.
     DBIx::Class::Storage::BlockRunner->new(
@@ -256,7 +258,7 @@ sub _dbic_block_runner {
         # never overrides the important ones below
         %{ $self->dbic_retry_opts },
 
-        storage  => $self->dbic_storage,
+        storage  => $storage,
         wrap_txn => ($method eq 'txn' ? 1 : 0),
     )->run($coderef);
 }
@@ -670,128 +672,24 @@ has max_id => (
     isa      => PositiveOrZeroInt,
 );
 
-#pod =head2 Private Attributes
+#pod =head3 loop_state
 #pod
-#pod =head3 _loop_state
-#pod
-#pod These variables exist solely for the processing loop.  They should be cleared out after
-#pod use.  Most of the complexity is needed for chunk resizing.
-#pod
-#pod =over
-#pod
-#pod =item timer
-#pod
-#pod Timer for debug messages.  Always spans the time between debug messages.
-#pod
-#pod =item start
-#pod
-#pod The real start ID that the loop is currently on.  May continue to exist within iterations
-#pod if chunk resizing is trying to find a valid range.  Otherwise, this value will become
-#pod undef when a chunk is finally processed.
-#pod
-#pod =item end
-#pod
-#pod The real end ID that the loop is currently looking at.  This is always redefined at the
-#pod beginning of the loop.
-#pod
-#pod =item prev_end
-#pod
-#pod Last "processed" value of C<end>.  This also includes skipped blocks.  Used in C<start>
-#pod calculations and to determine if the end of the loop has been reached.
-#pod
-#pod =item max_end
-#pod
-#pod The maximum ending ID.  This will be C<$DB_MAX_ID> if L</process_past_max> is set.
-#pod
-#pod =item last_range
-#pod
-#pod A hashref of keys used for the bisecting of one block.  Cleared out after a block has
-#pod been processed or skipped.
-#pod
-#pod =item last_timings
-#pod
-#pod An arrayref of hashrefs, containing data for the previous 5 runs.  This data is used for
-#pod runtime targeting.
-#pod
-#pod =item multiplier_range
-#pod
-#pod The range (in units of L</chunk_size>) between the start and end IDs.  This starts at 1
-#pod (at the beginning of the loop), but may expand or shrink depending on chunk count checks.
-#pod Resets after block processing.
-#pod
-#pod =item multiplier_step
-#pod
-#pod Determines how fast multiplier_range increases, so that chunk resizing happens at an
-#pod accelerated pace.  Speeds or slows depending on what kind of limits the chunk count
-#pod checks are hitting.  Resets after block processing.
-#pod
-#pod =item checked_count
-#pod
-#pod A check counter to make sure the chunk resizing isn't taking too long.  After ten checks,
-#pod it will give up, assuming the block is safe to process.
-#pod
-#pod =item chunk_size
-#pod
-#pod The I<current> chunk size, which might be adjusted by runtime targeting.
-#pod
-#pod =item chunk_count
-#pod
-#pod Records the results of the C<COUNT(*)> query for chunk resizing.
-#pod
-#pod =item prev_check
-#pod
-#pod A short string recording what happened during the last chunk resizing check.  Exists
-#pod purely for debugging purposes.
-#pod
-#pod =item prev_runtime
-#pod
-#pod The number of seconds the previously processed chunk took to run, not including sleep
-#pod time.
-#pod
-#pod =item progress_bar
-#pod
-#pod The progress bar being used in the loop.  This may be different than L</progress_bar>,
-#pod since it could be auto-generated.
-#pod
-#pod =back
+#pod A L<DBIx::BatchChunker::LoopState> object designed to hold variables during the
+#pod processing loop.  The object will be cleared out after use.  Most of the complexity is
+#pod needed for chunk resizing.
 #pod
 #pod =cut
 
-has _loop_state => (
+has loop_state => (
     is       => 'rw',
-    isa      => HashRef,
+    isa      => InstanceOf['DBIx::BatchChunker::LoopState'],
     required => 0,
     init_arg => undef,
-    lazy     => 1,
-    builder  => '_build_loop_state',
-    clearer  => '_clear_loop_state',
+    clearer  => 'clear_loop_state',
 );
 
-sub _build_loop_state {
-    my $self = shift;
-
-    my $start = $self->min_id;
-
-    return {
-        timer    => time,
-        start    => $start,
-        end      => $start + $self->chunk_size - 1,
-        prev_end => $start - 1,
-        max_end  => ($self->process_past_max ? $DB_MAX_ID : $self->max_id),
-
-        last_range       => {},
-        last_timings     => [],
-        multiplier_range => 0,
-        multiplier_step  => 1,
-        checked_count    => 0,
-        chunk_size       => $self->chunk_size,
-        chunk_count      => undef,
-        prev_check       => '',
-        prev_runtime     => undef,
-
-        progress_bar     => undef,
-    };
-}
+# Backwards-compatibility
+*_loop_state = \&loop_state;
 
 around BUILDARGS => sub {
     my $next  = shift;
@@ -877,9 +775,10 @@ around BUILDARGS => sub {
     );
 
     # Other sanity checks
-    die 'Range calculations requires one of these attr sets: rsc, rs, or dbi_connector|dbic_storage + min_stmt + max_stmt' unless (
+    die 'Range calculations require one of these attr sets: rsc, rs, or dbi_connector|dbic_storage + min_stmt + max_stmt' unless (
         defined $args{rsc} ||
-        (defined $args{min_stmt} && defined $args{max_stmt})
+        (defined $args{min_stmt} && defined $args{max_stmt}) ||
+        (!defined $args{dbi_connector} && !defined $args{dbic_storage} && defined $args{coderef})  # DIY mode is exempt
     );
 
     die 'Block execution requires one of these attr sets: dbi_connector|dbic_storage + stmt, rs + coderef, or coderef' unless (
@@ -1023,7 +922,7 @@ sub calculate_ranges {
     }
 
     # Set the ranges and return
-    return 0 unless defined $min_id && $max_id;
+    return 0 unless defined $min_id && defined $max_id;
 
     $self->min_id( int $min_id );
     $self->max_id( int $max_id );
@@ -1107,16 +1006,23 @@ sub execute {
     }
 
     # Loop state setup
-    $self->_clear_loop_state;
-    my $ls = $self->_loop_state;
-    $ls->{progress_bar} = $progress;
+    $self->clear_loop_state;
+    my $ls = $self->loop_state( DBIx::BatchChunker::LoopState->new({
+        batch_chunker => $self,
+        progress_bar  => $progress,
+    }) );
 
     # Da loop
-    while ($ls->{prev_end} < $ls->{max_end} || $ls->{start}) {
-        $ls->{multiplier_range} += $ls->{multiplier_step};
-        $ls->{start}             = $ls->{prev_end} + 1 unless defined $ls->{start};   # this could be already set because of early 'next' calls
-        $ls->{end}               = $ls->{start} + $ls->{multiplier_range} * $ls->{chunk_size} - 1;
-        $ls->{chunk_count}       = undef;
+    while ($ls->prev_end < $ls->max_end || $ls->start) {
+        $ls->multiplier_range($ls->multiplier_range + $ls->multiplier_step);
+        $ls->start           ($ls->prev_end + 1) unless defined $ls->start;   # this could be already set because of early 'next' calls
+        $ls->end(
+            min(
+                $ls->start + ceil($ls->multiplier_range * $ls->chunk_size) - 1, # ceil, because multiplier_* could be fractional
+                $ls->max_end, # ensure we never exceed max_end
+            )
+        );
+        $ls->chunk_count     (undef);
 
         next unless $self->_process_past_max_checker;
 
@@ -1124,7 +1030,7 @@ sub execute {
         next unless $self->_process_block;
 
         # Record the time quickly
-        $ls->{prev_runtime} = time - $ls->{timer};
+        $ls->prev_runtime(time - $ls->timer);
 
         # Give the DB a little bit of breathing room
         sleep $self->sleep if $self->sleep;
@@ -1134,16 +1040,9 @@ sub execute {
         $self->_runtime_checker;
 
         # End-of-loop activities (skipped by early next)
-        $ls->{start}     = undef;
-        $ls->{prev_end}  = $ls->{end};
-        $ls->{timer}     = time;
-
-        $ls->{last_range}       = {};
-        $ls->{multiplier_range} = 0;
-        $ls->{multiplier_step}  = 1;
-        $ls->{checked_count}    = 0;
+        $ls->_reset_chunk_state;
     }
-    $self->_clear_loop_state;
+    $self->clear_loop_state;
 
     # Keep the finished time from the progress bar, in case there are other loops or output
     unless ($progress->silent) {
@@ -1164,7 +1063,7 @@ sub execute {
 sub _process_block {
     my ($self) = @_;
 
-    my $ls      = $self->_loop_state;
+    my $ls      = $self->loop_state;
     my $conn    = $self->dbi_connector;
     my $coderef = $self->coderef;
     my $rs      = $self->rs;
@@ -1174,29 +1073,29 @@ sub _process_block {
     my $count_stmt = $self->count_stmt;
     if ($count_stmt && defined $self->dbic_storage) {
         $self->_dbic_block_runner( run => sub {
-            ($self->_loop_state->{chunk_count}) = $self->dbic_storage->dbh->selectrow_array(
+            $self->loop_state->chunk_count( $self->dbic_storage->dbh->selectrow_array(
                 @$count_stmt,
                 (@$count_stmt == 1 ? undef : ()),
-                @$ls{qw< start end >},
-            );
+                $ls->start, $ls->end,
+            ) );
         });
     }
     elsif ($count_stmt) {
-        ($ls->{chunk_count}) = $conn->run(sub {
+        $ls->chunk_count( $conn->run(sub {
             $_->selectrow_array(
                 @$count_stmt,
                 (@$count_stmt == 1 ? undef : ()),
-                @$ls{qw< start end >},
+                $ls->start, $ls->end,
             );
-        });
+        }) );
     }
     elsif (defined $rs) {
         $chunk_rs = $rs->search({
-            $self->id_name => { -between => [@$ls{qw< start end >}] },
+            $self->id_name => { -between => [$ls->start, $ls->end] },
         });
 
         $self->_dbic_block_runner( run => sub {
-            $self->_loop_state->{chunk_count} = $chunk_rs->count;
+            $self->loop_state->chunk_count( $chunk_rs->count );
         });
     }
 
@@ -1211,14 +1110,14 @@ sub _process_block {
         my @prepare_args = @$stmt > 2 ? @$stmt[0..1] : @$stmt;
         my @execute_args = (
             (@$stmt > 2 ? @$stmt[2..$#$stmt] : ()),
-            @$ls{qw< start end >},
+            $ls->start, $ls->end,
         );
 
         if ($self->single_rows && $coderef) {
             # Transactional work
             if ($self->dbic_storage) {
                 $self->_dbic_block_runner( txn => sub {
-                    $self->_loop_state->{timer} = time;  # reset timer on retries
+                    $self->loop_state->_mark_timer;  # reset timer on retries
 
                     my $sth = $self->dbic_storage->dbh->prepare(@prepare_args);
                     $sth->execute(@execute_args);
@@ -1228,7 +1127,7 @@ sub _process_block {
             }
             else {
                 $conn->txn(sub {
-                    $self->_loop_state->{timer} = time;  # reset timer on retries
+                    $self->loop_state->_mark_timer;  # reset timer on retries
 
                     my $sth = $_->prepare(@prepare_args);
                     $sth->execute(@execute_args);
@@ -1241,7 +1140,7 @@ sub _process_block {
             # Bulk work (or DML)
             if ($self->dbic_storage) {
                 $self->_dbic_block_runner( run => sub {
-                    $self->_loop_state->{timer} = time;  # reset timer on retries
+                    $self->loop_state->_mark_timer;  # reset timer on retries
 
                     my $sth = $self->dbic_storage->dbh->prepare(@prepare_args);
                     $sth->execute(@execute_args);
@@ -1251,7 +1150,7 @@ sub _process_block {
             }
             else {
                 $conn->run(sub {
-                    $self->_loop_state->{timer} = time;  # reset timer on retries
+                    $self->loop_state->_mark_timer;  # reset timer on retries
 
                     my $sth = $_->prepare(@prepare_args);
                     $sth->execute(@execute_args);
@@ -1268,7 +1167,7 @@ sub _process_block {
             # Transactional work
             $self->_dbic_block_runner( txn => sub {
                 # reset timer/$rs on retries
-                $self->_loop_state->{timer} = time;
+                $self->loop_state->_mark_timer;
                 $chunk_rs->reset;
 
                 while (my $row = $chunk_rs->next) { $self->coderef->($self, $row) }
@@ -1278,7 +1177,7 @@ sub _process_block {
             # Bulk work
             $self->_dbic_block_runner( run => sub {
                 # reset timer/$rs on retries
-                $self->_loop_state->{timer} = time;
+                $self->loop_state->_mark_timer;
                 $chunk_rs->reset;
 
                 $self->coderef->($self, $chunk_rs);
@@ -1288,7 +1187,7 @@ sub _process_block {
     else {
         ### Something a bit more free-form
 
-        $self->$coderef(@$ls{qw< start end >});
+        $self->$coderef($ls->start, $ls->end);
     }
 
     return 1;
@@ -1305,16 +1204,16 @@ sub _process_block {
 
 sub _process_past_max_checker {
     my ($self) = @_;
-    my $ls = $self->_loop_state;
-    my $progress = $ls->{progress_bar};
+    my $ls = $self->loop_state;
+    my $progress = $ls->progress_bar;
 
     return 1 unless $self->process_past_max;
-    return 1 unless $ls->{end} > $self->max_id;
+    return 1 unless $ls->end > $self->max_id;
 
     # No checks for DIY, if they didn't include a max_stmt
     unless (defined $self->rsc || $self->max_stmt) {
         # There's no way to size this, so skip past the max as one block
-        $ls->{end} = $ls->{max_end};
+        $ls->end($ls->max_end);
         return 1;
     }
 
@@ -1336,14 +1235,14 @@ sub _process_past_max_checker {
             $_->selectrow_array(@{ $self->max_stmt });
         });
     }
-    $ls->{timer} = time;  # the above query shouldn't impact runtimes
+    $ls->_mark_timer;  # the above query shouldn't impact runtimes
 
     if (!$new_max_id || $new_max_id eq '0E0') {
         # No max: No affected rows to change
         $progress->message('No max ID found; nothing left to process...') if $self->debug;
-        $ls->{end} = $ls->{max_end};
+        $ls->end($ls->max_end);
 
-        $ls->{prev_check} = 'no max';
+        $ls->prev_check('no max');
         return 0;
     }
     elsif ($new_max_id > $self->max_id) {
@@ -1356,14 +1255,17 @@ sub _process_past_max_checker {
     elsif ($new_max_id == $self->max_id) {
         # Same max ID
         $progress->message( sprintf 'Found max ID %u; same as end', $new_max_id ) if $self->debug;
-        $ls->{max_end} = $new_max_id;
+        $ls->max_end($new_max_id);
     }
     else {
         # Max too low
         $progress->message( sprintf 'Found max ID %u; ignoring...', $new_max_id ) if $self->debug;
-        $ls->{max_end} = $self->max_id;
+        $ls->max_end($self->max_id);
     }
 
+    # Run another boundary check with the new max_end value
+    $ls->end( min($ls->end, $ls->max_end) );
+
     return 1;
 }
 
@@ -1382,86 +1284,83 @@ sub _process_past_max_checker {
 
 sub _chunk_count_checker {
     my ($self) = @_;
-    my $ls = $self->_loop_state;
-    my $progress = $ls->{progress_bar};
+    my $ls = $self->loop_state;
+    my $progress = $ls->progress_bar;
 
     # Chunk sizing is essentially disabled, so bounce out of here
-    if ($self->min_chunk_percent <= 0 || !defined $ls->{chunk_count}) {
-        $ls->{prev_check} = 'disabled';
+    if ($self->min_chunk_percent <= 0 || !defined $ls->chunk_count) {
+        $ls->prev_check('disabled');
         return 1;
     }
 
-    my $chunk_percent = $ls->{chunk_count} / $ls->{chunk_size};
-    $ls->{checked_count}++;
+    my $chunk_percent = $ls->chunk_count / $ls->chunk_size;
+    $ls->checked_count( $ls->checked_count + 1 );
 
-    if    ($ls->{chunk_count} == 0 && $self->min_chunk_percent > 0) {
+    if    ($ls->chunk_count == 0 && $self->min_chunk_percent > 0) {
         # No rows: Skip the block entirely, and accelerate the stepping
         $self->_print_debug_status('skipped');
 
         $self->_increment_progress;
-        $ls->{start}     = undef;
-        $ls->{prev_end}  = $ls->{end};
-        $ls->{timer}     = time;
 
-        $ls->{last_range}       = {};
-        $ls->{multiplier_range} = 0;
-        $ls->{multiplier_step} *= 2;
-        $ls->{checked_count}    = 0;
+        my $step = $ls->multiplier_step;
+        $ls->_reset_chunk_state;
+        $ls->multiplier_step( $step * 2 );
 
-        $ls->{prev_check} = 'skipped rows';
+        $ls->prev_check('skipped rows');
         return 0;
     }
     elsif ($chunk_percent > 1 + $self->min_chunk_percent) {
         # Too many rows: Backtrack to the previous range and try to bisect
         $self->_print_debug_status('shrunk');
 
-        $ls->{timer} = time;
+        $ls->_mark_timer;
 
         # If we have a min/max range, bisect down the middle.  If not, walk back
         # to the previous range and decelerate the stepping, which should bring
         # it to a halfway point from this range and last.
-        $ls->{last_range}{max}  = $ls->{multiplier_range} if !defined $ls->{last_range}{max} || $ls->{multiplier_range} < $ls->{last_range}{max};
-        $ls->{multiplier_range} = $ls->{last_range}{min} || ($ls->{multiplier_range} - $ls->{multiplier_step});
-        $ls->{multiplier_step}  = int( defined $ls->{last_range}{min} ?
-            ($ls->{last_range}{max} - $ls->{last_range}{min}) / 2 :
-            $ls->{multiplier_step} / 2
+        my $lr = $ls->last_range;
+        $lr->{max} = $ls->multiplier_range if !defined $lr->{max} || $ls->multiplier_range < $lr->{max};
+        $ls->multiplier_range( $lr->{min} || ($ls->multiplier_range - $ls->multiplier_step) );
+        $ls->multiplier_step(
+            defined $lr->{min} ? ($lr->{max} - $lr->{min}) / 2 : $ls->multiplier_step / 2
         );
 
-        $ls->{prev_check} = 'too many rows';
+        $ls->prev_check('too many rows');
         return 0;
     }
 
     # The above two are more important than skipping the count checks.  Better to
     # have too few rows than too many.
 
-    elsif ($ls->{checked_count} > 10) {
+    elsif ($ls->checked_count > 10) {
         # Checked too many times: Just process it
-        $ls->{prev_check} = 'too many checks';
+        $ls->prev_check('too many checks');
         return 1;
     }
-    elsif ($ls->{end} >= $ls->{max_end}) {
+    elsif ($ls->end >= $ls->max_end) {
         # At the end: Just process it
-        $ls->{prev_check} = 'at max_end';
+        $ls->prev_check('at max_end');
         return 1;
     }
     elsif ($chunk_percent < $self->min_chunk_percent) {
         # Too few rows: Keep the start ID and accelerate towards a better endpoint
         $self->_print_debug_status('expanded');
 
-        $ls->{timer} = time;
+        $ls->_mark_timer;
 
         # If we have a min/max range, bisect down the middle.  If not, keep
         # accelerating the stepping.
-        $ls->{last_range}{min} = $ls->{multiplier_range} if !defined $ls->{last_range}{min} || $ls->{multiplier_range} > $ls->{last_range}{min};
-        $ls->{multiplier_step} = int( defined $ls->{last_range}{max} ?
-            ($ls->{last_range}{max} - $ls->{last_range}{min}) / 2 :
-            $ls->{multiplier_step} * 2
+        my $lr = $ls->last_range;
+        $lr->{min} = $ls->multiplier_range if !defined $lr->{min} || $ls->multiplier_range > $lr->{min};
+        $ls->multiplier_step(
+            defined $lr->{max} ? ($lr->{max} - $lr->{min}) / 2 : $ls->multiplier_step * 2
         );
-        $ls->{prev_check} = 'too few rows';
+
+        $ls->prev_check('too few rows');
         return 0;
     }
 
-    $ls->{prev_check} = 'nothing wrong';
+    $ls->prev_check('nothing wrong');
     return 1;
 }
 
@@ -1476,17 +1375,17 @@ sub _chunk_count_checker {
 
 sub _runtime_checker {
     my ($self) = @_;
-    my $ls = $self->_loop_state;
+    my $ls = $self->loop_state;
     return unless $self->target_time;
-    return unless $ls->{chunk_size} && $ls->{prev_runtime};  # prevent DIV/0
+    return unless $ls->chunk_size && $ls->prev_runtime;  # prevent DIV/0
 
-    my $timings = $ls->{last_timings};
+    my $timings = $ls->last_timings;
 
     my $new_timing = {
-        runtime     => $ls->{prev_runtime},
-        chunk_count => $ls->{chunk_count} || $ls->{chunk_size},
+        runtime     => $ls->prev_runtime,
+        chunk_count => $ls->chunk_count || $ls->chunk_size,
     };
-    $new_timing->{chunk_per} = $new_timing->{chunk_count} / $ls->{chunk_size};
+    $new_timing->{chunk_per} = $new_timing->{chunk_count} / $ls->chunk_size;
 
     # Rowtime: a measure of how much of the chunk_size actually impacted the runtime
     $new_timing->{rowtime} = $new_timing->{runtime} / $new_timing->{chunk_per};
@@ -1500,7 +1399,7 @@ sub _runtime_checker {
     my $avg_rowtime   = sum(map { $_->{rowtime} } @$timings) / $ttl;
     my $adjust_factor = $self->target_time / $avg_rowtime;
 
-    my $new_target_chunk_size = $ls->{chunk_size};
+    my $new_target_chunk_size = $ls->chunk_size;
     my $adjective;
     if    ($adjust_factor > 1.05) {
         # Too fast: Raise the chunk size
@@ -1514,10 +1413,10 @@ sub _runtime_checker {
     elsif ($adjust_factor < 0.95) {
         # Too slow: Lower the chunk size
 
-        return unless $ls->{prev_runtime} > $self->target_time;  # last runtime must actually be too high
+        return unless $ls->prev_runtime > $self->target_time;  # last runtime must actually be too high
 
         $new_target_chunk_size *=
-            ($ls->{prev_runtime} < $self->target_time * 3) ?
+            ($ls->prev_runtime < $self->target_time * 3) ?
             max(0.5, $adjust_factor) :  # never less than half...
             $adjust_factor              # ...unless the last runtime was waaaay off
         ;
@@ -1526,7 +1425,7 @@ sub _runtime_checker {
     }
 
     $new_target_chunk_size = int $new_target_chunk_size;
-    return if $new_target_chunk_size == $ls->{chunk_size};  # either nothing changed or it's too miniscule
+    return if $new_target_chunk_size == $ls->chunk_size;  # either nothing changed or it's too miniscule
     return if $new_target_chunk_size < 1;
 
     # Print out a debug line, if enabled
@@ -1539,14 +1438,14 @@ sub _runtime_checker {
             "Processing too %s, avg %4s of target time, adjusting chunk size from %s to %s",
             $adjective,
             $percent->format( 1 / $adjust_factor ),
-            $integer->format( $ls->{chunk_size} ),
+            $integer->format( $ls->chunk_size ),
             $integer->format( $new_target_chunk_size ),
         ) );
     }
 
     # Change it!
-    $ls->{chunk_size} = $new_target_chunk_size;
-    $ls->{last_timings} = [] if $adjective eq 'fast';  # never snowball too quickly
+    $ls->chunk_size($new_target_chunk_size);
+    $ls->_reset_last_timings if $adjective eq 'fast';  # never snowball too quickly
     return 1;
 }
 
@@ -1558,11 +1457,11 @@ sub _runtime_checker {
 
 sub _increment_progress {
     my ($self) = @_;
-    my $ls = $self->_loop_state;
-    my $progress = $ls->{progress_bar};
+    my $ls = $self->loop_state;
+    my $progress = $ls->progress_bar;
 
-    my $so_far = $ls->{end} - $self->min_id + 1;
-    $progress->target($so_far+1) if $ls->{end} > $self->max_id;
+    my $so_far = $ls->end - $self->min_id + 1;
+    $progress->target($so_far+1) if $ls->end > $self->max_id;
     $progress->update($so_far);
 }
 
@@ -1570,7 +1469,7 @@ sub _increment_progress {
 #pod
 #pod Prints out a standard debug status line, if debug is enabled.  What it prints is
 #pod generally uniform, but it depends on the processing action.  Most of the data is
-#pod pulled from L</_loop_state>.
+#pod pulled from L</loop_state>.
 #pod
 #pod =cut
 
@@ -1578,7 +1477,7 @@ sub _print_debug_status {
     my ($self, $action) = @_;
     return unless $self->debug;
 
-    my $ls    = $self->_loop_state;
+    my $ls    = $self->loop_state;
     my $sleep = $self->sleep || 0;
 
     # CLDR number formatters
@@ -1591,30 +1490,30 @@ sub _print_debug_status {
 
     my $message = sprintf(
         'IDs %6u to %6u %9s, %9s rows found',
-        $ls->{start}, $ls->{end}, $action,
-        $integer->format( $ls->{chunk_count} ),
+        $ls->start, $ls->end, $action,
+        $integer->format( $ls->chunk_count ),
     );
 
     $message .= sprintf(
         ' (%4s of chunk size)',
-        $percent->format( $ls->{chunk_count} / $ls->{chunk_size} ),
-    ) if $ls->{chunk_count};
+        $percent->format( $ls->chunk_count / $ls->chunk_size ),
+    ) if $ls->chunk_count;
 
     if ($action eq 'processed') {
         $message .= $sleep ?
             sprintf(
                 ', %5s+%s sec runtime+sleep',
-                $decimal->format( $ls->{prev_runtime} ),
+                $decimal->format( $ls->prev_runtime ),
                 $decimal->format( $sleep )
             ) :
             sprintf(
                 ', %5s sec runtime',
-                $decimal->format( $ls->{prev_runtime} ),
+                $decimal->format( $ls->prev_runtime ),
             )
         ;
     }
 
-    return $ls->{progress_bar}->message($message);
+    return $ls->progress_bar->message($message);
 }
 
 #pod =head1 SEE ALSO
@@ -1637,7 +1536,7 @@ DBIx::BatchChunker - Run large database changes safely
 
 =head1 VERSION
 
-version v0.940.1
+version v0.940.2
 
 =head1 SYNOPSIS
 
@@ -1771,6 +1670,11 @@ It's still valid to include L</min_stmt>, L</max_stmt>, and/or L</count_stmt> in
 constructor to enable features like L<max ID recalculation|/process_past_max> or
 L<chunk resizing|/min_chunk_percent>.
 
+If you're not going to include any min/max statements for L</calculate_ranges>, you will
+need to set L</min_id> and L</max_id> yourself, either in the constructor or before the
+L</execute> call.  Using L</construct_and_execute> is also not an option in this case, as
+this tries to call L</calculate_ranges> without a way to do so.
+
 =head3 TL;DR Version
 
     $stmt                             = Active DBI Processing
@@ -1806,11 +1710,10 @@ DBIC.  So far, there are two supported options:
     retry_handler = Coderef that returns true to continue to retry or false to re-throw
                     the last exception
 
-The default is to use DBIC's built-in retry options, the same way L<DBIx::Class::Storage::DBI/dbh_do>
-does it, which will retry once if the DB connection was disconnected.  If you specify any
-options, even a blank hashref, BatchChunker will fill in a default C<max_attempts> of 10,
-and an always-true C<retry_handler>.  This is similar to L<DBIx::Connector::Retry>'s
-defaults.
+The default is to let the DBIC storage engine handle its own protection, which will retry
+once if the DB connection was disconnected.  If you specify any options, even a blank
+hashref, BatchChunker will fill in a default C<max_attempts> of 10, and an always-true
+C<retry_handler>.  This is similar to L<DBIx::Connector::Retry>'s defaults.
 
 Under the hood, these are options that are passed to the as-yet-undocumented
 L<DBIx::Class::Storage::BlockRunner>.  The C<retry_handler> has access to the same
@@ -2034,90 +1937,11 @@ Manually setting this is not recommended, as each database is different and the
 information may have changed between the DB change development and deployment.  Instead,
 use L</calculate_ranges> to fill in these values right before running the loop.
 
-=head2 Private Attributes
+=head3 loop_state
 
-=head3 _loop_state
-
-These variables exist solely for the processing loop.  They should be cleared out after
-use.  Most of the complexity is needed for chunk resizing.
-
-=over
-
-=item timer
-
-Timer for debug messages.  Always spans the time between debug messages.
-
-=item start
-
-The real start ID that the loop is currently on.  May continue to exist within iterations
-if chunk resizing is trying to find a valid range.  Otherwise, this value will become
-undef when a chunk is finally processed.
-
-=item end
-
-The real end ID that the loop is currently looking at.  This is always redefined at the
-beginning of the loop.
-
-=item prev_end
-
-Last "processed" value of C<end>.  This also includes skipped blocks.  Used in C<start>
-calculations and to determine if the end of the loop has been reached.
-
-=item max_end
-
-The maximum ending ID.  This will be C<$DB_MAX_ID> if L</process_past_max> is set.
-
-=item last_range
-
-A hashref of keys used for the bisecting of one block.  Cleared out after a block has
-been processed or skipped.
-
-=item last_timings
-
-An arrayref of hashrefs, containing data for the previous 5 runs.  This data is used for
-runtime targeting.
-
-=item multiplier_range
-
-The range (in units of L</chunk_size>) between the start and end IDs.  This starts at 1
-(at the beginning of the loop), but may expand or shrink depending on chunk count checks.
-Resets after block processing.
-
-=item multiplier_step
-
-Determines how fast multiplier_range increases, so that chunk resizing happens at an
-accelerated pace.  Speeds or slows depending on what kind of limits the chunk count
-checks are hitting.  Resets after block processing.
-
-=item checked_count
-
-A check counter to make sure the chunk resizing isn't taking too long.  After ten checks,
-it will give up, assuming the block is safe to process.
-
-=item chunk_size
-
-The I<current> chunk size, which might be adjusted by runtime targeting.
-
-=item chunk_count
-
-Records the results of the C<COUNT(*)> query for chunk resizing.
-
-=item prev_check
-
-A short string recording what happened during the last chunk resizing check.  Exists
-purely for debugging purposes.
-
-=item prev_runtime
-
-The number of seconds the previously processed chunk took to run, not including sleep
-time.
-
-=item progress_bar
-
-The progress bar being used in the loop.  This may be different than L</progress_bar>,
-since it could be auto-generated.
-
-=back
+A L<DBIx::BatchChunker::LoopState> object designed to hold variables during the
+processing loop.  The object will be cleared out after use.  Most of the complexity is
+needed for chunk resizing.
 
 =head1 CONSTRUCTORS
 
@@ -2259,7 +2083,7 @@ Increments the progress bar.
 
 Prints out a standard debug status line, if debug is enabled.  What it prints is
 generally uniform, but it depends on the processing action.  Most of the data is
-pulled from L</_loop_state>.
+pulled from L</loop_state>.
 
 =head1 SEE ALSO
 
